/**
 * Copyright (c) 2021 Dominic Msters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "../libs.h"
#include "matrix.h"
#include "camera.h"
#include "texture.h"

/** Representation of a shader uniform */
typedef GLuint shaderuniform_t;

/**
 * Structure containing information about an OpenGL Shader. For simplicity sake
 * we demand certain uninforms to be present on the shader target.
 */
typedef struct {
  /** Pointer to an uploaded vertex shader program */
  shaderuniform_t shaderVertex;

  /** Pointer to an uploaded fragment shader program */
  shaderuniform_t shaderFrag;

  /** Pointer to an uploaded shader program linked */
  shaderuniform_t shaderProgram;
} shaderprogram_t;

/**
 * Compiles a shader from vertex and fragment shader code.
 * 
 * @param shader Shader to compile into.
 * @param vertexShaderSource The raw vertex shader code.
 * @param fragmentShaderSource The raw fragment shader code.
 */
void shaderInit(shaderprogram_t *shader,
  char *vertexShaderSource, char* fragmentShaderSource
);

/**
 * Cleanup and unload a previously loaded shader.
 * 
 * @param shader The shader to unload
 */
void shaderDispose(shaderprogram_t *shader);

/**
 * Attaches the supplied shader as the current shader.
 * 
 * @param shader The shader to attach
 */
void shaderUse(shaderprogram_t *shader);

/**
 * Attaches a camera to the shader.
 * 
 * @param shader Shader to attach to.
 * @param uniformView Uniform for the view matrix
 * @param uniformProjection Uniform for the projection matrix.
 * @param camera Camera to attach.
 */
void shaderUseCamera(
  shaderprogram_t *shader, shaderuniform_t uniformView,
  shaderuniform_t uniformProjection, camera_t *camera
);

/**
 * Attaches a texture to the shader.
 * 
 * @param shader Shader to attach to.
 * @param uniform Uniform to set the texture on to.
 * @param slot Which texture slot to bind to the uniform.
 */
void shaderUseTexture(
  shaderprogram_t *shader, shaderuniform_t uniform, textureslot_t slot
);

/**
 * Set's a specific shader uniform to a matrix.
 * 
 * @param shader Shader to apply to.
 * @param uniform Uniform on the shader to set.
 * @param matrix Matrix to apply.
 */
void shaderUseMatrix(
  shaderprogram_t *shader, shaderuniform_t uniform, matrix_t *matrix
);

/**
 * Set's the current translation matrix onto the shader for the next
 * render to use. Rotation order is set to YZX.
 * @param shader Shader to attach to.
 * @param uniform Uniform to set the position on to.
 * @param x X coordinate (world space).
 * @param y Y coordinate (world space).
 * @param z Z coordinate (world space).
 * @param pitch Pitch of the object (local space).
 * @param yaw Yaw of the object (local space).
 * @param roll Roll of the object (local space).
 */
void shaderUsePosition(shaderprogram_t *shader, shaderuniform_t uniform,
  float x, float y, float z,
  float pitch, float yaw, float roll
);


/**
 * Set's the current translation matrix onto the shader for the next
 * render to use. Also provides scaling controls.
 * 
 * @param shader Shader to attach to.
 * @param uniform Uniform to set the position on to.
 * @param x X coordinate (world space).
 * @param y Y coordinate (world space).
 * @param z Z coordinate (world space).
 * @param pitch Pitch of the object (local space).
 * @param yaw Yaw of the object (local space).
 * @param roll Roll of the object (local space).
 * @param scaleX X scale of model (1 being 100% scaled).
 * @param scaleY Y scale of model (1 being 100% scaled).
 * @param scaleZ Z scale of model (1 being 100% scaled).
 */
void shaderUsePositionAndScale(shaderprogram_t *shader, shaderuniform_t uniform,
  float x, float y, float z,
  float pitch, float yaw, float roll,
  float scaleX, float scaleY, float scaleZ
);

/**
 * Set a color on to the shader.
 * 
 * @param shader Shader to set the color on to.
 * @param uniform Uniform to set the color to.
 * @param color Color to set.
 */
void shaderUseColor(
  shaderprogram_t *shader, shaderuniform_t uniform, pixel_t color
);

/**
 * Locate a shader uniform by its name.
 * 
 * @param sp Shader program to get the uniform for.
 * @param name Name of the uniform to get.
 * @return The shader uniform.
 */
shaderuniform_t shaderGetUniformByName(shaderprogram_t *sp, const char *name);